package shenikan

import (
	"strings"
	"io/fs"

	// "cuelang.org/go/cue"
	"cuelang.org/go/cue/cuecontext"
	"cuelang.org/go/cue/load"

	"labprogramming.net/shenikandata"
)

type Glyph struct {
	Kind string `json:"kind"`
	Ortho string `json:"ortho"`
	Attrs []string `json:"attrs"`
}

type Replacement struct {
	Old string `json:"old"`
	New string `json:"new"`
}

type Dialect struct {
	Name string `json:"name"`
	Replacements []Replacement `json:"replacements"`
}

type Definition struct {
	PartOfSpeech string `json:"pos"`
	Short string `json:"short"`
	Long *string `json:"long"`
}

type DerivationalMorphology struct {
	Spelling string `json:"spelling"`
	Definitions []Definition
}

type PentaExtremes struct {
	I *string `json:"i"`
	U *string `json:"u"`
}

type DMPenta struct {
	Spelling string `json:"spelling"`
	Name *string `json:"name"`
	Extremes *PentaExtremes `json:"extremes"`
}

type Derivation struct {
	Root string `json:"root"`
	Via []string `json:"via"`
}

type Word struct {
	Spelling string `json:"spelling"`
	Definitions []Definition `json:"definitions"`
	Derivations []Derivation `json:"derivations"`
}

type Penta struct {
	Spelling string `json:"spelling"`
	Name *string `json:"name"`
	Extremes *PentaExtremes `json:"extremes"`
}

type IcosaExtremes struct {
	IX *string `json:"ix"`
	UX *string `json:"ux"`
	XI *string `json:"xi"`
	XU *string `json:"xu"`
}

type Icosa struct {
	Spelling string `json:"spelling"`
	Name *string `json:"name"`
	Extremes *IcosaExtremes `json:"extremes"`
}

type Dictionary struct {
	Glyphs []Glyph `json:"glyphs"`
	Dialects []Dialect `json:"dialects"`
	DMs []DerivationalMorphology `json:"dms"`
	DMPentas []DMPenta `json:"dmPentas"`
	Words []Word `json:"words"`
	Pentas []Penta `json:"pentas"`
	Icosas []Icosa `json:"icosas"`
}

type dictionaryData struct {
	Dict *Dictionary `json:"dictionary"`
}

var DictError error
var Dict *Dictionary = func() *Dictionary {
	var overlay map[string]load.Source = make(map[string]load.Source)

	sources, DictError := fs.Glob(shenikandata.Cues, "*")
	if DictError != nil {
		return nil
	}
	for source := range sources {
		content, DictError := shenikandata.Cues.ReadFile(sources[source])
		if DictError != nil {
			return nil
		}
		overlay["/.secretprefix/" + sources[source]] = load.FromBytes(content)
	}

	insts := load.Instances([]string{"."}, &load.Config{
		Dir: "/.secretprefix",
		Env: []string{},
		Overlay: overlay,
	})

	ctx := cuecontext.New()
	vals, DictError := ctx.BuildInstances(insts)
	if DictError != nil {
		return nil
	}

	var dd dictionaryData
	DictError = vals[0].Decode(&dd)

	return dd.Dict
}()

// Canonicalizes a string to unicode shenikan without pr or pl
func Canonicalize(shenikan string) (result string) {
	result = shenikan
	result = strings.ReplaceAll(result, "sh", "∫")
	result = strings.ReplaceAll(result, "th", "θ")
	result = strings.ReplaceAll(result, "`", "«")
	result = strings.ReplaceAll(result, "'", "»")
	result = strings.ReplaceAll(result, "c", "x")
	result = strings.ReplaceAll(result, "ç", "x")
	result = strings.ReplaceAll(result, "ç", "x")
	result = strings.ReplaceAll(result, "pr", "r")
	result = strings.ReplaceAll(result, "pl", "l")
	result = strings.ReplaceAll(result, "j", "p")
	result = strings.ReplaceAll(result, "g", "k")
	result = strings.ReplaceAll(result, "z", "s")
	result = strings.ReplaceAll(result, "d", "t")

	return
}

func RenderCSUR(shenikan string, offset int) (result string) {
	result = Canonicalize(shenikan)

	base := 0xF3A00 + 0x100 * offset
	var longest int = 0
	for _, glyph := range Dict.Glyphs {
		if len(glyph.Ortho) > longest {
			longest = len(glyph.Ortho)
		}
	}

	for length := longest; length > 0; length-- {
		for i, glyph := range Dict.Glyphs {
			if len(glyph.Ortho) == length {
				result = strings.ReplaceAll(result, glyph.Ortho, string(rune(base + i)))
			}
		}
	}

	return
}
